//
//	Software License Agreement (BSD License)
//	----------------------------------------
//	Copyright (c) 2013 IDEAL Software GmbH, Neuss, Germany. All rights reserved.
//                              www.IdealSoftware.com
//
//	Redistribution and use in source and binary forms, with or without modification,
//	are permitted provided that the following conditions are met:
//
//	* Redistributions of source code must retain the above copyright notice, this 
//	  list of conditions and the following disclaimer.
//
//	* Redistributions in binary form must reproduce the above copyright notice, this
//	  list of conditions and the following disclaimer in the documentation and/or
//	  other materials provided with the distribution.
//
//	* Neither the name IDEAL Software, nor the names of contributors
//	  may be used to endorse or promote products derived from this software without
//	  specific prior written permission of IDEAL Software.
//
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//	DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//	ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//
// Some comments on the Barcode Reader application:
// ------------------------------------------------
// - We experienced good barcode scanning results with a webcam that has an auto-focus. We used a HD 720p webcam  from Logitech.
// - The scanned barcodes must be illuminated brightly. If the barcode is scanned too dark or shadowed, it will not be recognized.
// - Good illumination and an auto-focus are very important! If the barcode image is not sharp, barcode recognition will fail.
//   You will notice that the webcam sometimes has difficulties to focus  barcodes. Give it some time and illuminate the
//   barcode brightly. Once the auto-focus works, scanning will work very good and very fast over a long series of different barcodes.
// - We included a PDF file (Barcode Samples.pdf) with example barcodes. The PDF file has been created using our product
//   Virtual Print Engine (VPE). Simply print the PDF and afterwards you can use it to test barcode recognition.
// - Unfortunately the HD resolution of the webcam  which would yield much better barcode recognition results  can not be 
//   used with OpenCV. At least not on Windows. When used, OpenCV will hang the second time the Barcode Reader application 
//   is executed. Beware: The process can not be terminated in this case  you need to hard-reboot!
//   Therefore the Barcode Reader application only uses 640 x 480 resolution, which already gives good barcode recognition 
//   results. We tried to use the VideoInputLibrary to retrieve images from the webcam, but it also hangs when using HD 
//   resolution. Later we found out that the VideoInputLibrary is used internally in the Windows version of OpenCV. 
//   If anyone can provide a better library to retrieve images from a webcam, please send us an e-mail.
// - When the Barcode Reader application can successfully decode a barcode, it beeps and displays the barcode content in the text window.
// - The live view of the webcam is shown in a smaller window, which can be scaled by code. This allows to embed it into a host application.
// - You can terminate the Barcode Reader application by pressing the ESCape key while the webcam window has the focus.


#include "../../opencv/include/opencv/cv.h" 
#include "../../opencv/include/opencv/highgui.h" 

#include <stdio.h>
#include <iostream>

#include "../../zxing/cpp/core/src/zxing/LuminanceSource.h"
#include "../../zxing/cpp/core/src/zxing/MultiFormatReader.h"
#include "../../zxing/cpp/core/src/zxing/oned/OneDReader.h"
#include "../../zxing/cpp/core/src/zxing/oned/EAN8Reader.h"
#include "../../zxing/cpp/core/src/zxing/oned/EAN13Reader.h"
#include "../../zxing/cpp/core/src/zxing/oned/Code128Reader.h"
#include "../../zxing/cpp/core/src/zxing/datamatrix/DataMatrixReader.h"
#include "../../zxing/cpp/core/src/zxing/qrcode/QRCodeReader.h"
#include "../../zxing/cpp/core/src/zxing/aztec/AztecReader.h"
#include "../../zxing/cpp/core/src/zxing/common/GlobalHistogramBinarizer.h"
#include "../../zxing/cpp/core/src/zxing/Exception.h"

using namespace std;
using namespace zxing;
using namespace oned;
using namespace datamatrix;
using namespace qrcode;
using namespace aztec;

typedef CvCapture *TDevice;


class OpenCVBitmapSource : public LuminanceSource
{
private:
	IplImage *m_pImage;

public:
	OpenCVBitmapSource(IplImage *image)
		: m_pImage(image)
	{
	}

	~OpenCVBitmapSource()
	{
		// when supporting cropping or rotation, the call to cvReleaseImage()
		// leads to crashes. obviously m_pImage must be managed ref-counted.
		// cvReleaseImage(&m_pImage);
	}

	int getWidth() const { return m_pImage->width; }
	int getHeight() const { return m_pImage->height; }

	unsigned char *getRow(int y, unsigned char *row)
	{
		if (!row)
			row = new unsigned char[getWidth()];

		memcpy(row, m_pImage->imageData + y * m_pImage->widthStep, getWidth());
		return row;
	}

	unsigned char *getMatrix()
	{
		int width = getWidth();
		int height =  getHeight();
		unsigned char *matrix = new unsigned char[width * height];
		for (int y = 0; y < height; y++)
			memcpy(matrix + y * width, m_pImage->imageData + y * m_pImage->widthStep, width);

		return matrix;
	}
	
	/*
	// The following methods are not supported by this demo (the DataMatrix Reader doesn't call these methods)
	bool isCropSupported() const { return false; }
	Ref<LuminanceSource> crop(int left, int top, int width, int height) {}
	bool isRotateSupported() const { return false; }
	Ref<LuminanceSource> rotateCounterClockwise() {}
	*/
};


void decode_image(Reader *reader, IplImage *image)
{
	try
	{
		Ref<OpenCVBitmapSource> source(new OpenCVBitmapSource(image));
		Ref<Binarizer> binarizer(new GlobalHistogramBinarizer(source));
		Ref<BinaryBitmap> bitmap(new BinaryBitmap(binarizer));
		Ref<Result> result(reader->decode(bitmap, DecodeHints(DecodeHints::DEFAULT_HINT)));
		cout << result->getText()->getText() << endl;
		Beep(440, 100);
	}
	catch (zxing::Exception& e)
	{
		// cerr << "Error: " << e.what() << endl;
	}
}


bool SetupCam(TDevice &device, int &width, int &height)
{
	IplImage *frame;
	int sel = 0;

#if 0	// ugly interface to select a CAMERA device
	int i;
	const int MAX_DEVICES = 8;
	int devices[MAX_DEVICES];
	int n = 0;
	for (i = 0; i < MAX_DEVICES; i++)
		devices[i] = -1;

	printf("select device:\n");
	for (i = 0; i < MAX_DEVICES; i++)
	{
		device = cvCaptureFromCAM(CV_CAP_DSHOW + i);
		if (device)
		{
			frame = cvQueryFrame(device);	// Get one frame
			if (frame)
			{
				printf("#%d - w = %d - h = %d\n", i, frame->width, frame->height);
				devices[n++] = i;
			}
			cvReleaseCapture(&device);
		}
	}

	if (devices[0] == -1)
	{
		fprintf( stderr, "ERROR: no video input devices found!\n" );
		getchar();
		return false;
	}

	do 
	{
		sel = getchar() - '0';
		for (i = 0; i < MAX_DEVICES && devices[i] != sel; i++)
			; // NOP
	}
	while (devices[i] != sel);
#endif

	// capture one frame from selected device
	device = cvCaptureFromCAM(CV_CAP_DSHOW + sel);

	// Setting the following resolution will make the process hang when run a second time.
	// The process then can not be killed !!!!
	// cvSetCaptureProperty(device, CV_CAP_PROP_FRAME_WIDTH, 1280);
	// cvSetCaptureProperty(device, CV_CAP_PROP_FRAME_HEIGHT, 720);

	frame = cvQueryFrame(device);	// Get one frame
	if (!frame)
	{
		fprintf( stderr, "ERROR: frame is null...\n" );
		getchar();
		return false;
	}

	width  = frame->width;
	height = frame->height;
	return true;
}


IplImage *CaptureFrame(TDevice &device)
{
	return cvQueryFrame(device);
}


void StopCam(TDevice &device)
{
	cvReleaseCapture(&device);
}


int main()
{
	TDevice device;
	IplImage *frame;
	int width, height;

	printf("Barcode Reader for Webcams\n\n"
		"Copyright (c) 2013 IDEAL Software GmbH, Neuss, Germany.\n"
		"www.idealsoftware.com\n\n"
		"Press ESC in the webcam window to terminate this application.\n\n");

	MultiFormatReader reader;
	if (!SetupCam(device, width, height))
		return 1;

	IplImage *gray = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 1);
	if (!gray)
	{
		fprintf(stderr, "ERROR: allocating image\n");
		getchar();
		return -1;
	}

	// create crop image
	int target_width = width * 2 / 3;
	int target_height = height * 2 / 3;
	int startx = (width - target_width) / 2;
	int starty = (height - target_height) / 2;
	IplImage *crop = cvCreateImage(cvSize(target_width, target_height), IPL_DEPTH_8U, 1);
	if (!crop)
	{
		fprintf(stderr, "ERROR: allocating image\n");
		getchar();
		return -1;
	}

	// Create a window in which the captured images will be presented
	cvNamedWindow("Barcode Reader", CV_WINDOW_AUTOSIZE);

	// Show the image captured from the camera in the window and repeat
	while (1)
	{
		// Get one frame
		frame = CaptureFrame(device);
		if (!frame)
		{
			fprintf(stderr, "ERROR: frame is null...\n");
			getchar();
			break;
		}

		// CV_RGB2GRAY: convert RGB image to grayscale
		cvCvtColor(frame, gray, CV_RGB2GRAY);

		// crop image
		cvSetImageROI(gray, cvRect(startx, starty, target_width, target_height));
		cvCopy(gray, crop, NULL);
		cvResetImageROI(gray);

#define DISPLAY_RESIZED
#ifdef DISPLAY_RESIZED
		// display resized original and paint crop rectangle
		float scale = 0.8f;
		IplImage *resized = cvCreateImage(cvSize((int)(width * scale), (int)(height * scale)), gray->depth, gray->nChannels);
		cvResize(gray, resized);

		// paint target crop rectangle
		int sstartx = (int)(startx * scale);
		int starget_width = (int)(target_width * scale);
		int sstarty = (int)(starty * scale);
		int starget_height = (int)(target_height * scale);
		for (int x = sstartx; x < sstartx + starget_width; x++)
		{
			resized->imageData[sstarty * resized->widthStep + x] = 0;
			resized->imageData[(sstarty + starget_height) * resized->widthStep + x] = 0;
		}

		for (int y = sstarty; y < sstarty + starget_height; y++)
		{
			resized->imageData[y * resized->widthStep + sstartx] = 0;
			resized->imageData[y * resized->widthStep + sstartx + starget_width] = 0;
		}

		cvShowImage("Barcode Reader", resized);
		cvReleaseImage(&resized);
#endif

		// Playing with contrast and sharpening, but it doesn't help much,
		// if the barcode is not illuminated brightly or if the webcam has no AUTO-FOCUS.
#if 1
		cv::Mat image;
		cv::Mat matGray(crop);
	#if 1
		// add contrast
		matGray.convertTo(image, -1, 2.0, 0);	// add contrast, keep brightness
	#else
		matGray.copyTo(image);
	#endif

	#if 1
		// sharpen image
		// use a gaussian smoothing filter and subtract the smoothed version from the original image
		cv::GaussianBlur(matGray, image, cv::Size(11, 11), 10, 10);
		double alpha = 5.0;		// alpha is the weight of the first array
		cv::addWeighted(matGray, alpha, image, 1.0 - alpha, 0, image);
	#else
		image.copyTo(matGray);
	#endif
		*crop = matGray;
#endif

#ifndef DISPLAY_RESIZED
		cvShowImage("Barcode Reader", crop);
#endif
		// decode barcode
		decode_image(&reader, crop);

		// Do not release the frame image!

		// If ESC key pressed, Key=0x10001B under OpenCV 0.9.7(linux version),
		// remove higher bits using AND operator:
		if ((cvWaitKey(10) & 255) == 27)
			break;
	}

	// Release the capture device and perform housekeeping
	StopCam(device);
	cvDestroyWindow("Barcode Reader");
	cvReleaseImage(&gray);
	cvReleaseImage(&crop);

	return 0;
}
